// Time 3 different versions of matrix vector multiplication
// parallelization: outer loop, inner loop, and both

#include <stdio.h>
#include <omp.h>
#include <stdlib.h> 
#include <sys/time.h>

// To measure timing
struct timeval tv; 
struct timezone tz; 
double time_start, time_end; 

void start_timer(){
  gettimeofday(&tv, &tz); 
  time_start = (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0; 
}

double get_wall_time(){
  gettimeofday(&tv, &tz); 
  time_end = (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0;
  double elapsed = time_end - time_start;
  return elapsed;
}

void zero_array(double *x, int n){
  for(int i=0; i<n; i++){
    x[i] = 0.0;
  }
}
#define PRINTING 0

void print_array(double *x, int n, char *prefix){
  if(PRINTING){
    for(int i=0; i<n; i++){
      printf("%s %d %f\n",prefix,i,x[i]);
    }
  }
}


int main(int argc, char **argv) { 
  if(argc < 3){
    printf("usage: matrix_vector_multiply <rows> <cols> <print?> [threads]\n");
    printf("  rows cols: ints, size of the matrix to use\n");
    printf("  threads: how many threads to use, to a limit more gets better performance\n");
    return -1;
  }

  int rows = atoi(argv[1]);
  int cols = atoi(argv[2]);

  if(argc > 3){
    omp_set_num_threads(atoi(argv[2]));
  }

  int i,j;

  // Raw matrix as a 1D array, initialize
  double *data = malloc(rows * cols * sizeof(double));
  if(data == NULL){
    printf("Insufficient memory\n");
    return -1;
  }
  for(i=0; i<rows*cols; i++){
    data[i] = i;
  }

  // set up row pointers
  double **matrix = malloc(rows * sizeof(double*));
  for(i=0; i<rows; i++){
    matrix[i] = data + i*cols;
  }

  // Allocate and initialize vector and result
  double *vector = malloc(cols * sizeof(double));
  for(i=0; i<cols; i++){
    vector[i] = i;
  }
  double *result = malloc(rows * sizeof(double));

  zero_array(result,rows);
  // Outer for loop multiplication
  start_timer();
  #pragma omp parallel for
  for(int i=0; i<rows; i++){
    if(i==0 && j==0){
      printf("#threads = %d (outer)\n",omp_get_num_threads());
    }        
    for(int j=0; j<cols; j++){
      result[i] += matrix[i][j] * vector[j];
    }
  }
  double time_outer_for = get_wall_time();
  print_array(result,rows,"O");

  zero_array(result,rows);
  // Inner for loop multiplication
  start_timer();
  for(int i=0; i<rows; i++){
    double sum = 0.0;
    #pragma omp parallel reduction(+:sum)
    {
      #pragma omp for
      for(int j=0; j<cols; j++){
        if(i==0 && j==0){
          printf("#threads = %d (inner)\n",omp_get_num_threads());
        }        
        sum += matrix[i][j] * vector[j];
      }
    }
    result[i] = sum;
      
  }
  double time_inner_for = get_wall_time();
  print_array(result,rows,"I");

  zero_array(result,rows);
  // Outer and Inner for loop multiplication
  start_timer();
  #pragma omp parallel for
  for(int i=0; i<rows; i++){
    if(i==0){
      printf("#threads = %d (both outer)\n",omp_get_num_threads());
    }
    double sum = 0.0;
    #pragma omp parallel reduction(+:sum)
    {
      #pragma omp for
      for(int j=0; j<cols; j++){
        if(i==0 && j==0){
          printf("#threads = %d (both inner) \n",omp_get_num_threads());
        }        
        sum += matrix[i][j] * vector[j];
      }
    }
    sum = result[i];
  }
  double time_both_for = get_wall_time();
  print_array(result,rows,"B");

  // Report times
  printf("Outer : %8.4f\n",time_outer_for);
  printf("Inner : %8.4f\n",time_inner_for);
  printf("Both  : %8.4f\n",time_both_for);

  free(data);
  free(matrix);
  free(vector);
  free(result);
}
